@@ -21,6 +21,8 @@
  * Author: JP Rosevear
  */
 
+#include <config.h>
+#include <netdb.h>
 #include <gnome.h>
 #include <glade/glade.h>
 #include "debug.h"
@@ -36,11 +38,13 @@
 
 /* Structs for game info */
 typedef struct {
-	GameInfo info;
+	EngineInfo info;
 	Engine *engine;
 	int game;
 	int ply;
-} ICSGameInfo;
+	gshort color;
+	gboolean flag;
+} ICSEngineInfo;
 
 struct _EngineIcsPrivate 
 {
@@ -98,16 +102,14 @@ static EngineClass *parent_class = NULL;
 
 static void class_init (EngineIcsClass *klass);
 static void init (EngineIcs *ics);
+static void destroy (GtkObject *obj);
 
 static GameView *ei_add_view (EngineIcs *engine);
-static void engine_ics_fill_menu (EngineNull *engine, GtkMenuShell *shell, gint pos);
-static void engine_ics_move (EngineNull *enull, gint from, gint to);
-static void engine_ics_destroy (GtkObject *obj);
-
-static void engine_ics_view_cb (GtkWidget *widget, gpointer data);
-static gboolean engine_ics_cb (GIOChannel *source, GIOCondition condition, gpointer data);
-static gboolean engine_ics_err_cb (GIOChannel *source, GIOCondition condition, gpointer data);
-static void engine_ics_command_cb (GtkWidget *widget, gchar *command, gpointer data);
+static void ei_view_cb (GtkWidget *widget, gpointer data);
+static gboolean ei_input_cb (GIOChannel *source, GIOCondition condition, gpointer data);
+static gboolean ei_err_cb (GIOChannel *source, GIOCondition condition, gpointer data);
+static void ei_resize_cb (GtkWidget *widget, GtkRequisition *requisition, gpointer data);
+static void ei_command_cb (GtkWidget *widget, gchar *command, gpointer data);
 
 /* Lex declarations */
 extern int yylex (EngineIcs *engine);
@@ -142,16 +144,11 @@ static void
 class_init (EngineIcsClass *klass)
 {
 	GtkObjectClass *object_class;
-	EngineClass *engine_class;
 	
 	object_class = GTK_OBJECT_CLASS (klass);
-	engine_class = ENGINE_CLASS (klass);
 	parent_class = gtk_type_class (engine_get_type ());
 
-	object_class->destroy = engine_ics_destroy;
-	
-	engine_class->engine_fill_menu = (EngineFillMenu)engine_ics_fill_menu;
-	engine_class->engine_move = (EngineMove)engine_ics_move;
+	object_class->destroy = destroy;
 }
 
 
@@ -165,8 +162,11 @@ init (EngineIcs *engine)
 	engine->priv = priv;
 
 	priv->term = SERVER_TERM (server_term_new ());
+	gtk_signal_connect_after (GTK_OBJECT (priv->term), "size_allocate",
+				  (GtkSignalFunc) ei_resize_cb,
+				  engine);
 	gtk_signal_connect (GTK_OBJECT (priv->term), "command",
-			    (GtkSignalFunc) engine_ics_command_cb,
+			    (GtkSignalFunc) ei_command_cb,
 			    engine);
 	
 	/* Set defaults */
@@ -188,6 +188,8 @@ engine_ics_new (const char *host,
 	EngineIcs *engine;
 	EngineIcsPrivate *priv;
 	GameView *view;
+	struct hostent *he;
+	char *host_addr;
 	char *arg[4];
 	
 	engine = gtk_type_new (engine_ics_get_type ());
@@ -205,8 +207,20 @@ engine_ics_new (const char *host,
 		arg[0] = g_strdup (prefs_get_telnetprogram ());
 	else
 		arg[0] = g_strdup (telnetprogram);
-
-	arg[1] = g_strdup (host);
+	
+	he = gethostbyname (host);
+	if (he->h_addr_list[0] != NULL) {
+		host_addr = g_strdup_printf ("%d.%d.%d.%d", 
+					     he->h_addr_list[0][0],
+					     he->h_addr_list[0][1],
+  					     he->h_addr_list[0][2],
+					     he->h_addr_list[0][3]);
+		arg[1] = host_addr;
+	} else {
+		arg[1] = g_strdup (host);
+	}
+	
+	
 	arg[2] = g_strdup (port);
 	arg[3] = NULL;
 		
@@ -216,8 +230,10 @@ engine_ics_new (const char *host,
 		     &priv->write_chan,
 		     &priv->childpid);
 
-	priv->read_cb = g_io_add_watch (priv->read_chan, G_IO_IN, engine_ics_cb, engine);
-	priv->err_cb = g_io_add_watch (priv->read_chan, G_IO_HUP, engine_ics_err_cb, engine);
+	priv->read_cb = g_io_add_watch (priv->read_chan, G_IO_IN, 
+					ei_input_cb, engine);
+	priv->err_cb = g_io_add_watch (priv->read_chan, G_IO_HUP | G_IO_ERR, 
+				       ei_err_cb, engine);
 
 	g_free (arg[0]);
 	g_free (arg[1]);
@@ -227,7 +243,23 @@ engine_ics_new (const char *host,
 	board_window_add_info (main_board_window, host, 
 			       GTK_WIDGET (priv->term));
 	
-	return GTK_OBJECT(engine);
+	return GTK_OBJECT (engine);
+}
+
+static void
+ei_resize (EngineIcs *engine)
+{
+	EngineIcsPrivate *priv;
+	guint width, height;
+	
+	priv = engine->priv;
+
+	if (priv->write_chan == NULL || !priv->loggedin)
+		return;
+	
+	server_term_get_size (priv->term, &width, &height);
+	write_child (priv->write_chan, "set width %d\n", width);
+	write_child (priv->write_chan, "set height %d\n", height);
 }
 
 static void
@@ -310,8 +342,51 @@ engine_ics_fill_menu (EngineNull *enull, GtkMenuShell *shell, gint pos)
 	gnome_app_fill_menu (shell, ei_menu, gtk_accel_group_get_default (), TRUE, pos);
 }
 
+static ICSEngineInfo *
+ei_get_info (GameView *view)
+{
+	return gtk_object_get_data (GTK_OBJECT (view), "gameinfo");
+}
+
+static ICSEngineInfo *
+ei_get_info_from_enull (EngineIcs *engine, EngineNull *enull)
+{
+	EngineIcsPrivate *priv;
+	GList *l;
+	
+	priv = engine->priv;
+	
+	for (l = priv->list; l != NULL; l = l->next) {
+		GameView *view = GAME_VIEW (l->data);
+		ICSEngineInfo *info  = ei_get_info (view);
+
+		if (info->engine == ENGINE (enull))
+			return gtk_object_get_data (GTK_OBJECT (view), 
+						    "gameinfo");
+	}	
+
+	return NULL;
+}
+
+static void
+ei_init_info (ICSEngineInfo *info) 
+{
+	g_free (info->info.white);
+	g_free (info->info.black);
+	g_free (info->info.description);
+	
+	info->info.white = NULL;
+	info->info.black = NULL;
+	info->info.description = g_strdup ("New board");
+
+	info->game = 0;
+	info->ply  = 0;
+	info->color = NONE;
+	info->flag = FALSE;
+}
+
 static void 
-engine_ics_move (EngineNull *enull, gint from, gint to)
+engine_ics_move (EngineNull *enull, Square from, Square to)
 {
 	EngineIcs *engine;
 	EngineIcsPrivate *priv;
@@ -324,26 +399,43 @@ engine_ics_move (EngineNull *enull, gint from, gint to)
 	write_child (priv->write_chan, "%s\n", str);
 }
 
+static void
+engine_ics_expire (EngineNull *enull, gshort color)
+{
+	EngineIcs *engine;
+	EngineIcsPrivate *priv;
+	ICSEngineInfo *icsinfo;
+	
+	engine = gtk_object_get_data (GTK_OBJECT (enull), "ics");
+	priv = engine->priv;
+
+	icsinfo = ei_get_info_from_enull (engine, enull);
+	if (icsinfo->flag && icsinfo->color != color && prefs_get_autoflag ()) {
+		debug_print (DEBUG_NORMAL, "Autoflagging\n");
+		write_child (priv->write_chan, "flag\n");
+	}
+}
+	
 static GameView *
 ei_add_view (EngineIcs *engine)
 {
 	EngineIcsPrivate *priv;
 	GtkWidget *view;
-	ICSGameInfo *info;
+	ICSEngineInfo *info;
 
 	priv = engine->priv;
 	
-	info = g_new0 (ICSGameInfo, 1);
-	info->info.description = g_strdup ("New board");
+	info = g_new0 (ICSEngineInfo, 1);
+	ei_init_info (info);
 	info->engine = ENGINE (engine_null_new ());
-	info->game = 0;
-	info->ply  = 0;
-	
+
 	/* Null engine needs info to handle callbacks */
 	engine_null_set_fill_menu (ENGINE_NULL (info->engine), 
-				   (EngineFillMenu)engine_ics_fill_menu);
+				   (EngineFillMenu) engine_ics_fill_menu);
 	engine_null_set_move (ENGINE_NULL (info->engine),
-			      (EngineMove)engine_ics_move);
+			      (EngineMove) engine_ics_move);
+	engine_null_set_expire (ENGINE_NULL (info->engine),
+				(EngineExpire) engine_ics_expire);
 
 	gtk_object_set_data (GTK_OBJECT (info->engine), "ics", engine);
 
@@ -356,43 +448,40 @@ ei_add_view (EngineIcs *engine)
 	/* Add it to our list */
 	priv->list = g_list_prepend (priv->list, view);
 	gtk_signal_connect (GTK_OBJECT (view), "destroy",
-			    (GtkSignalFunc) engine_ics_view_cb,
+			    (GtkSignalFunc) ei_view_cb,
 			    engine);
 	
 	return GAME_VIEW (view);
 }
 
-static
-ICSGameInfo *ei_get_info (EngineIcs *engine, GameView *view)
-{
-	return gtk_object_get_data (GTK_OBJECT (view), "gameinfo");
-}
-
 static GameView *
 ei_get_view (EngineIcs *engine, int game)
 {
 	EngineIcsPrivate *priv;
-	GtkObject *view;
-	GtkObject *alt_view;
-	ICSGameInfo *info;
+	GameView *view, *alt_view = NULL;
+	ICSEngineInfo *info;
 	GList *l;
 	
 	priv = engine->priv;
 	
-	alt_view = NULL;
-	
 	for (l = priv->list; l != NULL; l = l->next) {		
-		view = GTK_OBJECT (l->data);
-		info = ei_get_info (engine, GAME_VIEW (view));
+		view = l->data;
+		info = ei_get_info (view);
 
 		if (info->game == game)
-			return GAME_VIEW (view);
+			return view;
 		else if (info->game == 0)
 			alt_view = view;
 	}	
 
-	if (alt_view)
-		return GAME_VIEW (alt_view);
+	if (alt_view != NULL) {
+		ICSEngineInfo *info;
+		
+		info = ei_get_info (alt_view);
+		ei_init_info (info);
+
+		return alt_view;
+	}
 	
 	return ei_add_view (engine);
 }
@@ -445,16 +534,17 @@ engine_ics_invalidpassword (EngineIcs *engine)
 }
 
 void 
-engine_ics_prompt(EngineIcs *engine, gchar *str) 
+engine_ics_prompt (EngineIcs *engine, gchar *str) 
 {
 	EngineIcsPrivate *priv;
-
+	
 	priv = engine->priv;
 
 	if (!priv->loggedin) {
+		priv->loggedin = TRUE;
 		server_term_set_prompt (priv->term, str);
 		write_child (priv->write_chan, "set style 12\n");
-		priv->loggedin = TRUE;
+		ei_resize (engine);
 	}
 }
 
@@ -462,29 +552,32 @@ void
 engine_ics_result (EngineIcs *engine, char *result)
 {
 	GameView *view;
-	ICSGameInfo *icsinfo;
+	ICSEngineInfo *icsinfo;
 	char *rstr;
 	gint resultval, game;
 
 	/* Figure out what game */
-	rstr = strchr(result, ' ');
+	rstr = strchr (result, ' ');
 	rstr++;
-	game = atoi(rstr); 
+	game = atoi (rstr); 
 
 	/* Figure out what happened */
 	rstr = strrchr(result, ' ');
 	rstr++;
-	if (!strcmp(rstr, "1-0"))
+	if (!strcmp (rstr, "1-0"))
 		resultval = GAME_WHITE;
-	else if(!strcmp(rstr, "0-1"))
+	else if (!strcmp (rstr, "0-1"))
 		resultval = GAME_BLACK;
-	else
+	else if (!strcmp (rstr, "1/2-1/2"))
 		resultval = GAME_DRAW;
+	else
+		resultval = GAME_ABORTED;
 
 	/* Update the info struct */
 	view = ei_get_view (engine, game);
 
-	icsinfo = ei_get_info (engine, GAME_VIEW (view));	
+	icsinfo = ei_get_info (view);
+	icsinfo->game = 0;
 	icsinfo->info.result = resultval;
 
 	/* Signal the new info */
@@ -495,21 +588,15 @@ engine_ics_result (EngineIcs *engine, char *result)
 int 
 engine_ics_input (char *buf, int max) 
 {
-	char b[1024];
 	guint len;
 
-	g_io_channel_read (chan, b, sizeof (b), &len);
-
-	if (len > 0)
-		memcpy (buf, b, len);
-	else if (len == 0)
-		debug_print(DEBUG_NORMAL, "Zero length string read.");
+	g_io_channel_read (chan, buf, max, &len);
 
 	return len;
 }
 
 void 
-engine_ics_output(EngineIcs *engine, char *str) 
+engine_ics_output (EngineIcs *engine, char *str) 
 {
 	EngineIcsPrivate *priv;
 	
@@ -521,13 +608,13 @@ engine_ics_output(EngineIcs *engine, char *str)
 static Position *
 engine_ics_board_to_pos (char *b)
 {
-	int i,j;
+	int i, j;
 	Position *pos;
 
 	pos = POSITION (position_new ());
 	for (i = 0; i < 8; i++) {
 		for (j = 0; j < 8; j++) {
-			int n= A8 - i * 10 + j;                                                                             
+			Square n = A8 - i * 10 + j;                                                                             
 			switch (b[(8*i)+j]) {
 			case 'R':
 				pos->square[n] = WR;
@@ -583,11 +670,11 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 	GameView *view;
 	Board *board;
 	Position *pos;
-	ICSGameInfo *icsinfo;
+	ICSEngineInfo *icsinfo;
 	gchar **tokens;
 	char b[65], m[10];
 	int game, colour, ply, rel;
-	int from,to;
+	Square from, to;
 	int i;
 
 	priv = engine->priv;
@@ -605,7 +692,7 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 	view = ei_get_view (engine, game);
 	board = game_view_get_board (view);
 
-	icsinfo = ei_get_info (engine, GAME_VIEW (view));
+	icsinfo = ei_get_info (view);
 	
 	/* Find the user's relation to the game */
 	rel = atoi (tokens[RELATION]);
@@ -620,16 +707,20 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 	}
 
 	/* Get the last move */
-	if (!strcmp(tokens[VERBOSE], "o-o") ||
-	    !strcmp(tokens[VERBOSE], "o-o-o")) {
+	if (!strcmp (tokens[VERBOSE], "o-o") ||
+	    !strcmp (tokens[VERBOSE], "o-o-o")) {
 		strncpy (m, tokens[VERBOSE], strlen (tokens[VERBOSE])+1);
-	} else if (strcmp(tokens[VERBOSE], "none")) {
-		strncpy (m, tokens[VERBOSE], 2);
+	} else if (strcmp (tokens[VERBOSE], "none")) {
 		m[0] = tokens[VERBOSE][2];
 		m[1] = tokens[VERBOSE][3];
 		m[2] = tokens[VERBOSE][5];
 		m[3] = tokens[VERBOSE][6];
-		m[4] = '\0';
+		if (tokens[VERBOSE][7] == '=') {
+			m[4] = tokens[VERBOSE][8];
+			m[5] = '\0';
+		} else {
+			m[4] = '\0';
+		}
 	} else {
 		strcpy (m, "none");
 	}
@@ -641,24 +732,23 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 		colour = BLACK;
 		
 	/* Figure out what ply this is */
-	if (!strcmp(m, "none")) {
+	if (!strcmp (m, "none"))
 		ply =  0;
-	} else if (colour == WHITE) {
-		ply = atoi (tokens[MOVE_NUM])*2-2;
-	} else {
-		ply = atoi (tokens[MOVE_NUM])*2-1;
-	}
+	else if (colour == WHITE)
+		ply = atoi (tokens [MOVE_NUM])*2-2;
+	else
+		ply = atoi (tokens [MOVE_NUM])*2-1;
 
 	/* Update the info structure */
 	if (!icsinfo->info.white ||
-	    !strcmp(icsinfo->info.white, tokens[WHITE_NAME]))
-		icsinfo->info.white = g_strdup (tokens[WHITE_NAME]);
+	    !strcmp (icsinfo->info.white, tokens [WHITE_NAME]))
+		icsinfo->info.white = g_strdup (tokens [WHITE_NAME]);
 	if (!icsinfo->info.black ||
-	    !strcmp(icsinfo->info.black, tokens[BLACK_NAME]))
-		icsinfo->info.black = g_strdup (tokens[BLACK_NAME]);
+	    !strcmp (icsinfo->info.black, tokens [BLACK_NAME]))
+		icsinfo->info.black = g_strdup (tokens [BLACK_NAME]);
 
-	icsinfo->info.white_time = atoi (tokens[WHITE_TIME]);
-	icsinfo->info.black_time = atoi (tokens[BLACK_TIME]);
+	icsinfo->info.white_time = atoi (tokens [WHITE_TIME]);
+	icsinfo->info.black_time = atoi (tokens [BLACK_TIME]);
 
 	icsinfo->info.tomove = colour;
 
@@ -669,11 +759,19 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 	
 	if (icsinfo->info.description)
 		g_free (icsinfo->info.description);
-	icsinfo->info.description = 
-		g_strdup_printf ("%s %s game, %s to move ", 
-				 tokens[START_TIME], tokens[START_INC],
-				 colour == WHITE ? "white" : "black");
+
+	icsinfo->info.description = g_strdup_printf ("%s %s game, %s to move ",
+						     tokens[START_TIME], 
+						     tokens[START_INC],
+						     colour == WHITE ? 
+						     "white" : "black");
 	
+	icsinfo->flag = (rel == 1 || rel == -1) ? TRUE : FALSE;
+	if (rel == 1)
+		icsinfo->color = colour;
+	else if (rel == -1)
+		icsinfo->color = (colour == WHITE) ? BLACK : WHITE;
+
 	g_strfreev (tokens);
 
 	/* Setup the board and figure out the move if there */
@@ -681,11 +779,10 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 	pos = board_get_position (board);
 	if (strcmp (m, "none") && ply - icsinfo->ply == 1) {
 		icsinfo->ply = ply;
-		if (rel != -1) {
-			ascii_to_move (pos, m, &from, &to);
+		ascii_to_move (pos, m, &from, &to);
+		if (position_get_color_to_move (pos) != colour)
 			gtk_signal_emit_by_name (GTK_OBJECT (icsinfo->engine),
 						 "move", from, to);
-		}
 	} else {
 		pos = engine_ics_board_to_pos (b);
 		position_set_color_to_move (pos, colour);
@@ -699,7 +796,7 @@ engine_ics_update_board (EngineIcs *engine, char *boardstring)
 }
 
 static void
-engine_ics_destroy (GtkObject *obj)
+destroy (GtkObject *obj)
 {
 	EngineIcs *engine;
 	EngineIcsPrivate *priv;
@@ -723,9 +820,9 @@ engine_ics_destroy (GtkObject *obj)
 }
 
 static gboolean
-engine_ics_cb (GIOChannel *source,
-	       GIOCondition condition,
- 	       gpointer data)
+ei_input_cb (GIOChannel *source,
+	     GIOCondition condition,
+	     gpointer data)
 {
 	int token;
 
@@ -736,17 +833,24 @@ engine_ics_cb (GIOChannel *source,
 }
 
 static gboolean
-engine_ics_err_cb (GIOChannel *source,
-		   GIOCondition condition,
-		   gpointer data)
+ei_err_cb (GIOChannel *source,
+	   GIOCondition condition,
+	   gpointer data)
 {
-	g_error ("ICS Engine connection died");	
-	
+	gtk_signal_emit_by_name (GTK_OBJECT (data), "error",
+				 ENGINE_ERROR_CRITICAL);
+
 	return FALSE;
 }
 
 static void
-engine_ics_command_cb (GtkWidget *widget, gchar *command, gpointer data)
+ei_resize_cb (GtkWidget *widget, GtkRequisition *requisition, gpointer data)
+{
+	ei_resize (data);
+}
+
+static void
+ei_command_cb (GtkWidget *widget, gchar *command, gpointer data)
 {
 	EngineIcs *engine;
 	EngineIcsPrivate *priv;
@@ -758,20 +862,20 @@ engine_ics_command_cb (GtkWidget *widget, gchar *command, gpointer data)
 }
 
 static void
-engine_ics_view_cb (GtkWidget *widget, gpointer data)
+ei_view_cb (GtkWidget *widget, gpointer data)
 {
 	EngineIcs *engine;
 	EngineIcsPrivate *priv;
 	GList *elem;
-	ICSGameInfo *info;
+	ICSEngineInfo *info;
 
 	engine = ENGINE_ICS (data);
 	priv = engine->priv;
 
-	elem = g_list_find (priv->list, (gpointer)widget);
-	if (elem) {
+	elem = g_list_find (priv->list, widget);
+	if (elem != NULL) {
 		/* Free the info */
-		info = gtk_object_get_data (GTK_OBJECT (elem->data), "gameinfo");
+		info = ei_get_info (GAME_VIEW (widget));
 		gtk_object_remove_data (GTK_OBJECT (elem->data), "gameinfo");
 		g_free (info);
 		
@@ -779,3 +883,4 @@ engine_ics_view_cb (GtkWidget *widget, gpointer data)
 	}
 }
 
+
